#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
/*
 * pptpserver.c
 *
 * starts the PPTP Linux Server
 *
 *
 * Copyright (C) 1999  Matthew Ramsay <matthewr@moreton.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifndef FALSE
#define FALSE 0
#define TRUE !FALSE
#endif

#include "inststr.h"
#include "pptpctrl.h"
#include "pptpdebug.h"
#include "pptpdefs.h"
#include "pptpgre.h"
#include "pptpserver.h"
#include "pty.h"



static int ctrl_establish_signal = 0;
static int localgiven = FALSE;
static int remotegiven = FALSE;
static char *localipstr = NULL;
static char *remoteipstr = NULL;
static int debuglevel = 0;
static int speedgiven = FALSE;
static char *speedstr = NULL;

void do_squat_all(int sig) {

    ctrl_establish_signal = 1;

}

// local function prototypes
int launch_pppd(char *ttydev);
int get_call_id(u_int16_t *call_id, u_int16_t *peer_call_id);

static void showusage(char *prog)
{
  printf("usage:\n %s [-hdvslr] [--help] [--debug] [--version]\n  [--speed 9999]  [--localip x.x.x.x] [--remoteip x.x.x.x]\n",prog);
}
static void showversion(char *prog)
{
  printf("%s version %s\n",prog,VERSION);
}

int main(int argc, char **argv, char **envp) {
	
	pid_t parent_pid, child_pid;
	
	char ptydev[PTYMAX], ttydev[TTYMAX];
	int pty_fd;
	
	struct in_addr inetaddr;
	
	u_int16_t call_id, peer_call_id;
		
	int pptp_ctrl_fd;
	
	int pipe_fd[2]; // interprocess communication file descriptors
	
	// debugging
        char debug_msg[MAX_DEBUG_MSG_SIZE];
        unsigned char inetfoo[4];
	char buf[32]; // for changing the processes name

	// command line options
	int digit_optind = 0;
	int c;


        ///////////////////////////////////////////////////////////////////
	while(1) {
		int this_option_optind  = optind ? optind : 1;
		int option_index = 0;
		static struct option long_options[] =
		{
			{"help", 0, 0, 0},
			{"version", 0, 0, 0},
			{"localip", 1, 0, 0},
			{"remoteip", 1, 0, 0},
			{"debug", 0, 0, 0},
			{"speed", 1, 0, 0},
			{0,0,0,0}
		};
		c = getopt_long (argc, argv, "hvl:r:ds:",
				long_options, &option_index);
		if (c == -1) break;
		if (c == 0) { /* convert long options to short form */
			switch(option_index) {
			case 0: c = 'h'; break;
			case 1: c = 'v'; break;
			case 2: c = 'l'; break;
			case 3: c = 'r'; break;
			case 4: c = 'd'; break;
			case 5: c = 's'; break;
			}
		}
		switch(c) {
		case 'h': { showusage(argv[0]); exit(0); break; }
		case 'v': { showversion(argv[0]); exit(0); break; }
		case 'd': { debuglevel++; break; }
		case 'l':
			{
			if (strlen(optarg) > 15) {
				printf("Invalid local IP address\n");
				exit(-1);
			}
			localgiven = TRUE;
			localipstr = strdup(optarg);
			break;
			}
		case 'r':
			{
			if (strlen(optarg) > 15) {
				printf("Invalid remote IP address\n");
				exit(-1);
			}
			remotegiven = TRUE;
			remoteipstr = strdup(optarg);
			break;
			}
		case 's':
			{
			speedgiven = TRUE;
			speedstr = strdup(optarg);
			break;
			}
		default: { showusage(argv[0]); exit(-1); break; }
		}

	}



        ///////////////////////////////////////////////////////////////////
	pptpd_debug("Hello!", VERBOSE);

		
        // STOLEN - Finds an open pty/tty pair.. this is for PPP
	pty_fd = getpseudotty(ttydev, ptydev);
		
	sprintf(debug_msg, "ttydev = %s", ttydev);
	pptpd_debug(debug_msg, VERBOSE);
	///////////////////////////////////////////////////////////////////
	
	
	if((pipe(pipe_fd)) == -1) {
	
		pptpd_debug("pipe() failed", CRITICAL);
		perror("creating pipe");
		
	}
	
	signal(SIGUSR1, do_squat_all);
	parent_pid = getpid();
	
	switch (child_pid = fork()) {
	case 0: // child
	
		signal(SIGUSR1, SIG_DFL);
		child_pid = getpid();
		
		// launch the CTRL connection
		pptp_ctrl_fd = pptp_open_ctrl_connection(argc, argv, envp, pipe_fd, parent_pid);
		///////////////////////////////////////////////////////////
		
		// Launch PPP daemon
		pptpd_debug("launching PPPD...", VERBOSE);
		launch_pppd(ttydev);
		///////////////////////////////////////////////////////////
		
		break;
		
	default: // parent
		
		pptpd_debug("GRE connection waiting for the lights to change", VERBOSE);
		
		/*
		 * There is still a very small race condition here.  If a signal
		 * occurs after signaled is checked but before pause is called,
		 * things will hang.
		 */
		if (!ctrl_establish_signal) {
		
			pause();
		 	
		}
		
		pptpd_debug("GRE FORK: about to read any info on the pipe", VERBOSE);
		close(pipe_fd[PIPE_WRITE]);
		read(pipe_fd[PIPE_READ], inetfoo, 4);
		close(pipe_fd[PIPE_READ]);
		
		memcpy(&inetaddr, inetfoo, 4);
		pptpd_byte_debug("GRE FORK: inetaddr of the client", inetfoo, 4);

				
		pptpd_debug("GRE connection got a green light!", VERBOSE);
	
		// change the process name to help debugging
		strcpy(buf, "PPTP: GRE-PPP/PPP-GRE Connection");
		inststr(argc, argv, envp, buf);
		
		if((get_call_id(&call_id, &peer_call_id)) == -1) {
	
			pptpd_debug("failed to assign call and peer call ID's", CRITICAL);
		
		}
		
		sprintf(debug_msg, "call ID = %x, peer call ID = %x", call_id, peer_call_id);
		pptpd_debug(debug_msg, VERBOSE);
		
	       	pptpd_debug("Launching GRE connection server..", VERBOSE);
       		pptp_gre_copy(call_id, peer_call_id, pty_fd, inetaddr);

       		pptpd_debug("GRE connection finished..", VERBOSE);
	
	}
	
	pptpd_debug("Adios", VERBOSE);
	
	return 0;
	
}
	

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// local functions only below
///////////////////////////////////////////////////////////////////////////

/*
 * lanuch_pppd
 *
 * Launches the PPP daemon. The PPP daemon is responsible for assigning the
 * PPTP client its IP address.. These values are assigned via the command
 * line.
 *
 * args: ttydev - a tty paired with a pty
 * retn: 0 on success, -1 on failure.
 *
 */
int launch_pppd(char *ttydev) {

	char *pppd_argv[5];
	
	char localIP[16], remoteIP[16];
	char pppInterfaceIPs[32];
	char connectSpeed[10];
	
	int result;
	int an = 0;
	
	// debugging
        char debug_msg[MAX_DEBUG_MSG_SIZE];

                        	
	pppd_argv[an] = PPPD_BINARY;
	an++;
	
	pppd_argv[an] = ttydev;
	an++;
	
	if (speedgiven) {
	
		strncpy(connectSpeed,speedstr,(sizeof(connectSpeed) - 1));
		pppd_argv[an] = connectSpeed;
		an++;
		
		if (debuglevel > 0) {
			sprintf(debug_msg, "PPPD: Connection speed = %s", connectSpeed);
			pptpd_debug(debug_msg, VERBOSE);
		}
		
	}
	
	if (localgiven) {
	
		strncpy(localIP,localipstr,(sizeof(localIP) - 1));
		
	}
	
	if (remotegiven) {
	
		strncpy(remoteIP,remoteipstr,(sizeof(remoteIP) - 1));
		
	}
	
	if(localgiven && remotegiven) {
		
		sprintf(pppInterfaceIPs, "%s:%s", localIP, remoteIP);
		strcpy(pppd_argv[an], pppInterfaceIPs);
		an++;

	} else if (localgiven && !remotegiven) {
	
		sprintf(pppInterfaceIPs, "%s:", localIP);
		strcpy(pppd_argv[an], pppInterfaceIPs);
		an++;
		
	} else if (remotegiven && !localgiven) {
	
		sprintf(pppInterfaceIPs, ":%s", remoteIP);
		strcpy(pppd_argv[an], pppInterfaceIPs);
		an++;
		
	}
	
	if(localgiven || remotegiven) {
		
		if(debuglevel > 0) {
			
			sprintf(debug_msg, "PPP local/remote interface IP addresses = %s", pppInterfaceIPs);	
			pptpd_debug(debug_msg, VERBOSE);
			
		}
		
	}
	
	// must end with NULL or the pppd daemon spits chips
	pppd_argv[an] = NULL;
	an++;

	if((execvp(pppd_argv[0], pppd_argv)) == -1) {
	
		pptpd_debug("Failed to launch PPP daemon..", CRITICAL);
		return -1;
	
	}
	
	return 0;

}


/*
 * get_call_id
 *
 * Assigns a call ID and peer call ID to the session.
 *
 * args: call_id (OUT) - the call ID for the session
 *	 peer_call_id (OUT) - the peer call ID for the session
 * retn: 0 on success, -1 on failure
 *
 */
int get_call_id(u_int16_t *call_id, u_int16_t *peer_call_id) {
	
	// @FIXME.. this is pathetic
	*call_id = 0x0;
	*peer_call_id = 0x1;
	
	return 0;

}
